/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmQaIO.cxx
/// @brief  Module for ROOT objects IO interface (implementation)
/// @author S.Zharko <s.zharko@gsi.de>
/// @since  29.03.2023

#include "CbmQaIO.h"

#include "TFile.h"

// ---------------------------------------------------------------------------------------------------------------------
//
CbmQaIO::CbmQaIO(TString prefixName, std::shared_ptr<ObjList_t> pObjList) : fsPrefix(prefixName), fpvObjList(pObjList)
{
  if (!fsPrefix.IsNull()) { fsPrefix += "_"; }
  if (!fpvObjList.get()) { fpvObjList = std::make_shared<ObjList_t>(); }
}

// ---------------------------------------------------------------------------------------------------------------------
//
CbmQaIO::~CbmQaIO() {}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmQaIO::SetHistoProperties(TH1* pHist) { pHist->SetStats(true); }

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmQaIO::SetCanvasProperties(TCanvas* pCanv)
{
  constexpr double left   = 0.15;
  constexpr double bottom = 0.15;
  constexpr double right  = 0.10;
  constexpr double top    = 0.10;
  pCanv->SetMargin(left, right, bottom, top);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmQaIO::SetConfigName(const char* path)
{
  fsConfigName = path;
  try {
    fConfigNode = YAML::LoadFile(path)["qa"][fsPrefix.Data()];
  }
  catch (const YAML::BadFile& exc) {
    std::stringstream msg;
    msg << "configuration file for QA \"" << path << "\" does not exist";
    throw std::runtime_error(msg.str());
  }
  catch (const YAML::ParserException& exc) {
    std::stringstream msg;
    msg << "configuration file for QA \"" << path << "\" is improperly formatted";
    throw std::runtime_error(msg.str());
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmQaIO::WriteToFile(TFile* pOutFile) const
{
  LOG(info) << "CbmQaIO: Writing objects in file \033[1;31m" << pOutFile->GetName() << "\033[0m";
  pOutFile->cd();
  for (auto& [pObject, sPath] : (*fpvObjList)) {
    if (!pOutFile->GetDirectory(sPath)) { pOutFile->mkdir(sPath); }
    pOutFile->cd(sPath);
    if (pObject) { pObject->Write(); }
  }
}
